# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np
import qtawesome as qta

from PySide6 import QtCore
from PySide6.QtWidgets import QPushButton, QVBoxLayout, QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._base import AlgoBase
from acconeer.exptool.a121.algo._plugins import (
    DetectorBackendPluginBase,
    DetectorPlotPluginBase,
    DetectorViewPluginBase,
)
from acconeer.exptool.app.new import (
    BUTTON_ICON_COLOR,
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GridGroupBox,
    Message,
    MiscErrorView,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    VerticalGroupBox,
    is_task,
    pidgets,
)

from ._detector import Detector, DetectorConfig, DetectorResult, _load_algo_data


@attrs.mutable(kw_only=True)
class PlotConfig(AlgoBase):
    number_of_zones: Optional[int] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    plot_config: PlotConfig = attrs.field(factory=PlotConfig)


class PluginPresetId(Enum):
    DEFAULT = auto()


class BackendPlugin(DetectorBackendPluginBase[SharedState]):

    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: DetectorConfig()
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.plot_config = PlotConfig.from_json(file["plot_config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast(sync=True)

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast(sync=True)

    @is_task
    def update_plot_config(self, *, config: PlotConfig) -> None:
        self.shared_state.plot_config = config
        self.broadcast(sync=True)

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        _create_h5_string_dataset(file, "plot_config", self.shared_state.plot_config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast(sync=True)

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            detector_config=self.shared_state.config,
        )
        self._detector_instance.start(recorder)
        self.callback(
            GeneralMessage(
                name="setup",
                kwargs=dict(
                    detector_config=self.shared_state.config,
                    sensor_config=Detector._get_sensor_config(self.shared_state.config),
                    plot_config=self.shared_state.plot_config,
                    estimated_frame_rate=self._detector_instance.estimated_frame_rate,
                ),
                recipient="plot_plugin",
            )
        )

    def end_session(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        self._detector_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._detector_instance is None:
            raise RuntimeError
        result = self._detector_instance.get_next()

        self._frame_count += 1

        self.callback(GeneralMessage(name="rate_stats", data=self.client._rate_stats))
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))
        self.callback(GeneralMessage(name="plot", data=result, recipient="plot_plugin"))


class PlotPlugin(DetectorPlotPluginBase):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup_from_message(self, message: GeneralMessage) -> None:
        assert message.kwargs is not None
        self.setup(**message.kwargs)

    def update_from_message(self, message: GeneralMessage) -> None:
        assert isinstance(message.data, DetectorResult)
        self.update(message.data)

    def setup(
        self,
        detector_config: DetectorConfig,
        sensor_config: a121.SensorConfig,
        plot_config: PlotConfig,
        estimated_frame_rate: float,
    ) -> None:
        self.detector_config = detector_config
        self.distances = np.linspace(
            detector_config.start_m, detector_config.end_m, sensor_config.num_points
        )

        self.history_length_s = 5
        self.history_length_n = int(round(self.history_length_s * estimated_frame_rate))
        self.intra_history = np.zeros(self.history_length_n)
        self.inter_history = np.zeros(self.history_length_n)

        if plot_config.number_of_zones:
            self.num_sectors = min(plot_config.number_of_zones, self.distances.size)
            self.sector_size = max(1, -(-self.distances.size // self.num_sectors))
        else:
            max_num_of_sectors = max(6, self.distances.size // 3)
            self.sector_size = max(1, -(-self.distances.size // max_num_of_sectors))
            self.num_sectors = -(-self.distances.size // self.sector_size)

        self.sector_offset = (self.num_sectors * self.sector_size - self.distances.size) // 2
        win = self.plot_layout

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=0,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Distance (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_plot.setVisible(False)
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)

        # Depthwise presence plot

        self.move_plot = pg.PlotItem(title="Depthwise presence")
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Distance (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        self.move_plot.setXRange(self.distances[0], self.distances[-1])
        self.intra_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(1))
        if not self.detector_config.intra_enable:
            self.intra_curve.hide()

        self.inter_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(0))
        if not self.detector_config.inter_enable:
            self.inter_curve.hide()

        self.move_smooth_max = et.utils.SmoothMax(
            self.detector_config.frame_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_plot.addItem(self.present_text_item)
        self.move_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=1,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.detector_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            intra_dashed_pen = pg.mkPen(intra_color, width=2.5, style=QtCore.Qt.DashLine)
            intra_pen = pg.mkPen(intra_color, width=2)
        else:
            intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.detector_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=1,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.detector_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            inter_dashed_pen = pg.mkPen(inter_color, width=2.5, style=QtCore.Qt.DashLine)
            inter_pen = pg.mkPen(inter_color, width=2)
        else:
            inter_pen = et.utils.pg_pen_cycler(0)
            inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.detector_config.inter_detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=2, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_plot, row=0, col=0)
        sublayout.addItem(self.sector_plot, row=0, col=1)

    def update(self, data: DetectorResult) -> None:
        noise = data.processor_extra_result.lp_noise
        self.noise_curve.setData(self.distances, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data.presence_distance

        self.inter_curve.setData(self.distances, data.inter_depthwise_scores)
        self.intra_curve.setData(self.distances, data.intra_depthwise_scores)
        m = self.move_smooth_max.update(
            np.max(np.maximum(data.inter_depthwise_scores, data.intra_depthwise_scores))
        )
        m = max(
            m,
            2
            * np.maximum(
                self.detector_config.intra_detection_threshold,
                self.detector_config.inter_detection_threshold,
            ),
        )
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data.presence_detected))

        self.set_present_text_y_pos(m)

        if data.presence_detected:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        # Intra presence

        move_hist_xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.intra_history = np.roll(self.intra_history, -1)
        self.intra_history[-1] = data.intra_presence_score

        m_hist = max(
            float(np.max(self.intra_history)),
            self.detector_config.intra_detection_threshold * 1.05,
        )
        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(move_hist_xs, self.intra_history)

        # Inter presence

        self.inter_history = np.roll(self.inter_history, -1)
        self.inter_history[-1] = data.inter_presence_score

        m_hist = max(
            float(np.max(self.inter_history)),
            self.detector_config.inter_detection_threshold * 1.05,
        )
        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(move_hist_xs, self.inter_history)

        # Sector

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data.presence_detected:
            index = (
                data.processor_extra_result.presence_distance_index + self.sector_offset
            ) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y: float) -> None:
        x_pos = self.distances[0] + (self.distances[-1] - self.distances[0]) / 2
        self.present_text_item.setPos(x_pos, 0.95 * y)
        self.not_present_text_item.setPos(x_pos, 0.95 * y)


class ViewPlugin(DetectorViewPluginBase):
    def __init__(self, app_model: AppModel, view_widget: QWidget) -> None:
        super().__init__(app_model=app_model, view_widget=view_widget)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(
            qta.icon("fa5s.play-circle", color=BUTTON_ICON_COLOR),
            "Start measurement",
            self.sticky_widget,
        )
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(
            qta.icon("fa5s.stop-circle", color=BUTTON_ICON_COLOR),
            "Stop",
            self.sticky_widget,
        )
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.defaults_button = QPushButton(
            qta.icon("mdi6.restore", color=BUTTON_ICON_COLOR),
            "Reset settings and calibrations",
            self.sticky_widget,
        )
        self.defaults_button.clicked.connect(self._send_defaults_request)

        button_group = GridGroupBox("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.defaults_button, 1, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = VerticalGroupBox("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdParameterWidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_parameter_changed.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor[DetectorConfig](
            title="Detector parameters",
            factory_mapping=self._get_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.plot_config_editor = AttrsConfigEditor[PlotConfig](
            title="Plot parameters",
            factory_mapping={
                "number_of_zones": pidgets.OptionalIntParameterWidgetFactory(
                    name_label_text="Detection zones",
                    checkbox_label_text="Override",
                    limits=(1, 10),
                    init_set_value=3,
                )
            },
            parent=self.scrolly_widget,
        )
        self.plot_config_editor.sig_update.connect(self._on_plot_config_update)
        scrolly_layout.addWidget(self.plot_config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "start_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range start",
                suffix=" m",
                decimals=3,
            ),
            "end_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range end",
                suffix=" m",
                decimals=3,
            ),
            "profile": pidgets.OptionalEnumParameterWidgetFactory(
                name_label_text="Profile",
                checkbox_label_text="Override",
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (shortest)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (longest)",
                },
            ),
            "step_length": pidgets.OptionalIntParameterWidgetFactory(
                name_label_text="Step length",
                checkbox_label_text="Override",
                limits=(1, None),
                init_set_value=24,
            ),
            "frame_rate": pidgets.FloatParameterWidgetFactory(
                name_label_text="Frame rate",
                suffix=" Hz",
                decimals=1,
                limits=(1, 100),
            ),
            "sweeps_per_frame": pidgets.IntParameterWidgetFactory(
                name_label_text="Sweeps per frame",
                limits=(1, 4095),
            ),
            "hwaas": pidgets.IntParameterWidgetFactory(
                name_label_text="HWAAS",
                limits=(1, 511),
            ),
            "inter_frame_idle_state": pidgets.EnumParameterWidgetFactory(
                enum_type=a121.IdleState,
                name_label_text="Inter frame idle state",
                label_mapping={
                    a121.IdleState.DEEP_SLEEP: "Deep sleep",
                    a121.IdleState.SLEEP: "Sleep",
                    a121.IdleState.READY: "Ready",
                },
            ),
            "intra_enable": pidgets.CheckboxParameterWidgetFactory(
                name_label_text="Enable fast motion detection"
            ),
            "intra_detection_threshold": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Intra detection threshold",
                decimals=2,
                limits=(0, 5),
            ),
            "intra_frame_time_const": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Intra time constant",
                suffix=" s",
                decimals=2,
                limits=(0, 1),
            ),
            "intra_output_time_const": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Intra output time constant",
                suffix=" s",
                decimals=2,
                limits=(0.01, 20),
                log_scale=True,
            ),
            "inter_enable": pidgets.CheckboxParameterWidgetFactory(
                name_label_text="Enable slow motion detection"
            ),
            "inter_phase_boost": pidgets.CheckboxParameterWidgetFactory(
                name_label_text="Enable phase boost"
            ),
            "inter_detection_threshold": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Inter detection threshold",
                decimals=2,
                limits=(0, 5),
            ),
            "inter_frame_fast_cutoff": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Inter fast cutoff freq.",
                suffix=" Hz",
                decimals=2,
                limits=(1, 50),
                log_scale=True,
            ),
            "inter_frame_slow_cutoff": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Inter slow cutoff freq.",
                suffix=" Hz",
                decimals=2,
                limits=(0.01, 1),
                log_scale=True,
            ),
            "inter_frame_deviation_time_const": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Inter time constant",
                suffix=" s",
                decimals=2,
                limits=(0.01, 20),
                log_scale=True,
            ),
            "inter_output_time_const": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Inter output time constant",
                suffix=" s",
                decimals=2,
                limits=(0.01, 20),
                log_scale=True,
            ),
            "inter_frame_presence_timeout": pidgets.OptionalIntParameterWidgetFactory(
                name_label_text="Presence timeout",
                checkbox_label_text="Enable",
                suffix=" s",
                limits=(1, 30),
                init_set_value=5,
            ),
        }

    def on_backend_state_update(self, backend_plugin_state: Optional[SharedState]) -> None:
        if backend_plugin_state is not None and backend_plugin_state.config is not None:
            results = backend_plugin_state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        state = app_model.backend_plugin_state

        if state is None:
            self.start_button.setEnabled(False)
            self.stop_button.setEnabled(False)
            self.defaults_button.setEnabled(False)

            self.config_editor.set_data(None)
            self.config_editor.setEnabled(False)
            self.plot_config_editor.set_data(None)
            self.plot_config_editor.setEnabled(False)
            self.sensor_id_pidget.set_selected_sensor(None, [])

            return

        assert isinstance(state, SharedState)

        self.defaults_button.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)

        self.config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.config_editor.set_data(state.config)
        self.plot_config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.plot_config_editor.set_data(state.plot_config)
        self.sensor_id_pidget.set_selected_sensor(state.sensor_id, app_model.connected_sensors)
        self.sensor_id_pidget.setEnabled(app_model.plugin_state.is_steady)

        self.start_button.setEnabled(app_model.is_ready_for_session())
        self.stop_button.setEnabled(app_model.plugin_state == PluginState.LOADED_BUSY)

    def _on_config_update(self, config: DetectorConfig) -> None:
        self.app_model.put_backend_plugin_task("update_config", {"config": config})

    def _on_plot_config_update(self, config: PlotConfig) -> None:
        self.app_model.put_backend_plugin_task("update_plot_config", {"config": config})

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "sync":
            self._log.debug(f"{type(self).__name__} syncing")

            self.config_editor.sync()
        else:
            raise RuntimeError("Unknown message")

    def _send_defaults_request(self) -> None:
        self.app_model.put_backend_plugin_task("restore_defaults")

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        self.app_model.put_backend_plugin_task("update_sensor_id", {"sensor_id": sensor_id})


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


PRESENCE_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="presence_detector",
    title="Presence detector",
    description="Detect human presence.",
    family=PluginFamily.DETECTOR,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

import acconeer.exptool as et
from acconeer.exptool import a121


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    session_config = a121.SessionConfig(
        [
            {
                1: a121.SensorConfig(
                    subsweeps=[
                        a121.SubsweepConfig(
                            start_point=25,
                            step_length=2,
                            num_points=30,
                            profile=a121.Profile.PROFILE_1,
                            hwaas=10,
                        ),
                        a121.SubsweepConfig(
                            start_point=75,
                            step_length=4,
                            num_points=25,
                            profile=a121.Profile.PROFILE_3,
                            hwaas=20,
                        ),
                    ],
                )
            },
            {
                1: a121.SensorConfig(
                    receiver_gain=20,
                ),
            },
        ],
        extended=True,
    )

    extended_metadata = client.setup_session(session_config)

    pg_updater = PGUpdater(session_config, extended_metadata)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        extended_result = client.get_next()

        try:
            pg_process.put_data(extended_result)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.disconnect()


class PGUpdater:
    def __init__(
        self,
        session_config: a121.SessionConfig,
        extended_metadata: list[dict[int, a121.Metadata]],
    ) -> None:
        self.session_config = session_config
        self.extended_metadata = extended_metadata

    def setup(self, win):
        self.all_plots = []
        self.all_curves = []
        self.all_smooth_maxs = []

        for group_idx, group in enumerate(self.session_config.groups):
            group_plots = {}
            group_curves = {}
            group_smooth_maxs = {}

            for sensor_id, sensor_config in group.items():
                title = f"Group {group_idx} / Sensor {sensor_id}"
                plot = win.addPlot(title=title)
                plot.setMenuEnabled(False)
                plot.setMouseEnabled(x=False, y=False)
                plot.hideButtons()
                plot.showGrid(x=True, y=True)

                plot.setLabel("bottom", "Depth (m)")
                plot.setLabel("left", "Amplitude")

                curves = []
                for i in range(sensor_config.num_subsweeps):
                    curve = plot.plot(pen=et.utils.pg_pen_cycler(i))
                    curves.append(curve)

                group_plots[sensor_id] = plot
                group_curves[sensor_id] = curves

                smooth_max = et.utils.SmoothMax(self.session_config.update_rate)
                group_smooth_maxs[sensor_id] = smooth_max

            self.all_plots.append(group_plots)
            self.all_curves.append(group_curves)
            self.all_smooth_maxs.append(group_smooth_maxs)

    def update(self, extended_result: list[dict[int, a121.Result]]) -> None:
        for group_idx, group in enumerate(extended_result):
            for sensor_id, result in group.items():
                plot = self.all_plots[group_idx][sensor_id]
                curves = self.all_curves[group_idx][sensor_id]

                max_ = 0

                for sub_idx, subframe in enumerate(result.subframes):
                    x = get_distances_m(
                        self.session_config.groups[group_idx][sensor_id].subsweeps[sub_idx]
                    )
                    y = np.abs(subframe).mean(axis=0)
                    curves[sub_idx].setData(x, y)

                    max_ = max(max_, np.max(y))

                smooth_max = self.all_smooth_maxs[group_idx][sensor_id]
                plot.setYRange(0, smooth_max.update(max_))


def get_distances_m(config):
    range_p = np.arange(config.num_points) * config.step_length + config.start_point
    return range_p * 2.5e-3


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .client_info import ClientInfo
from .metadata import Metadata
from .record import PersistentRecord, Record, RecordException
from .result import Result, ResultContext
from .sensor_calibration import SensorCalibration
from .server_info import SensorInfo, ServerInfo
from .server_log_message import ServerLogMessage
from .stacked_results import StackedResults
from .utils import (
    complex_array_to_int16_complex,
    get_subsweeps_from_frame,
    int16_complex_array_to_complex,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

from acconeer.exptool.a121._core.entities import SensorCalibration, SessionConfig

from ._latest import ExplorationProtocolError
from ._no_5_2_mhz import ExplorationProtocol_No_5_2MHz_PRF


class ExplorationProtocol_NoCalibrationReuse(ExplorationProtocol_No_5_2MHz_PRF):
    @classmethod
    def setup_command(
        cls,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> bytes:
        if calibrations:
            raise ExplorationProtocolError(
                "Connected Exploration server does not support "
                + "calibration reuse (passing 'calibrations')"
            )
        return super().setup_command(session_config, None)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import copy

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ThresholdMethod,
    calculate_bg_noise_std,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    # Define sensor configuration.
    sensor_config = a121.SensorConfig(
        subsweeps=[
            a121.SubsweepConfig(
                start_point=50,
                step_length=4,
                num_points=50,
                profile=a121.Profile.PROFILE_1,
                hwaas=32,
                phase_enhancement=True,
            )
        ],
        sweeps_per_frame=1,
    )

    # Calibrate noise.
    noise_sensor_config = copy.deepcopy(sensor_config)
    for subsweep in noise_sensor_config.subsweeps:
        # Disable Tx when measuring background noise.
        subsweep.enable_tx = False

    metadata = client.setup_session(noise_sensor_config)
    client.start_session()
    result = client.get_next()
    client.stop_session()

    stds = [
        calculate_bg_noise_std(subframe, subsweep_config)
        for (subframe, subsweep_config) in zip(result.subframes, noise_sensor_config.subsweeps)
    ]

    # Create processor for distance estimation.
    distance_context = ProcessorContext(
        bg_noise_std=stds,
    )
    distance_config = ProcessorConfig(
        processor_mode=ProcessorMode.DISTANCE_ESTIMATION,
        threshold_method=ThresholdMethod.CFAR,
        threshold_sensitivity=0.8,
    )
    distance_processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=distance_config,
        context=distance_context,
    )

    pg_updater = PGUpdater()
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    metadata = client.setup_session(sensor_config)
    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        extended_result = client.get_next()
        processed_data = distance_processor.process(extended_result)
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.disconnect()


class PGUpdater:
    def __init__(self):
        self.history = [np.NaN] * 100

    def setup(self, win):
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance(m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        legends = ["Sweep", "Threshold"]
        self.curves = {}
        for i, legend in enumerate(legends):
            pen = et.utils.pg_pen_cycler(i)
            brush = et.utils.pg_brush_cycler(i)
            symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
            feat_kw = dict(pen=pen, **symbol_kw)
            self.curves[legend] = self.sweep_plot.plot(**feat_kw, name=legends[i])

        self.smooth_max = et.utils.SmoothMax()

        self.dist_history_plot = win.addPlot(row=1, col=0)
        self.dist_history_plot.setMenuEnabled(False)
        self.dist_history_plot.showGrid(x=True, y=True)
        self.dist_history_plot.addLegend()
        self.dist_history_plot.setLabel("left", "Estimated distance (m)")
        self.dist_history_plot.setLabel("bottom", "History (frame)")
        self.dist_history_plot.addItem(pg.PlotDataItem())
        self.dist_history_plot.setXRange(0, len(self.history))

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.dist_history_curve = self.dist_history_plot.plot(**feat_kw)

    def update(self, d):
        sweep = d.extra_result.abs_sweep
        threshold = d.extra_result.used_threshold
        distances = d.extra_result.distances_m

        self.history.pop(0)
        if len(d.estimated_distances) != 0:
            self.history.append(d.estimated_distances[0])
        else:
            self.history.append(np.nan)

        self.curves["Sweep"].setData(distances, sweep)
        self.curves["Threshold"].setData(distances, threshold)
        self.sweep_plot.setYRange(
            0,
            self.smooth_max.update(np.amax(np.concatenate((sweep, threshold)))),
        )

        if not np.all(np.isnan(self.history)):
            self.dist_history_curve.setData(self.history)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
import warnings
from typing import Any, Optional, Union

from acconeer.exptool.a121._core import utils

from .sensor_config import SensorConfig
from .validation_error import ValidationError, ValidationResult, ValidationWarning


@utils.no_dynamic_member_creation
class SessionConfig:
    """Session configuration

    The session configuration defines the highest level of configuration available in Exploration
    Tool. It mainly consists of one or several sensor configurations (:class:`SensorConfig`) for
    the server to run. It also sets the update rate for the server, and the data format
    ("extended") returned from the client.

    Mapping from sensor ID to sensor config is done here in the session config. For example, if you
    want to want to use sensor 2, you can do:

    .. code-block:: python

        SessionConfig({2: SensorConfig(start_point=123)})

    The default sensor ID is 1. Going further, you may run multiple sensors at the same time (in
    parallel) like this:

    .. code-block:: python

        SessionConfig(
            {
                2: SensorConfig(start_point=123),
                3: SensorConfig(start_point=456),
            }
        )

    The dictionary shown above forms a *group* of sensor configs. Further extending upon this, you
    may specify multiple groups which are run in sequence, like this:

    .. code-block:: python

        SessionConfig(
            [
                {
                    2: SensorConfig(start_point=123),
                    3: SensorConfig(start_point=456),
                },
                {
                    2: SensorConfig(start_point=789),
                },
            ]
        )

    You may reuse the same sensor across groups. If a sensor is used multiple times, a
    reconfiguration will be done prior to each measurement.

    A session config with multiple sensor configs (in a single group or multiple) is considered
    "extended". This is reflected in the shape of some return types. A SessionConfig with a single
    SensorConfig is not extended, but the return values can be passed as extended with the keyword
    argument ``extended=True``.

    :param arg: The sensor configuration(s).
    :param extended:
        Forces whether to use the extended format or not. If not given (``= None``), the extended
        format will be used automatically if multiple sensor configs are given.
    :param update_rate:
        The update rate limit on the server. Defaults to None, not limiting the rate.
    :raises ValueError: If the session config must be extended but ``extended=False``.
    """

    _groups: list[dict[int, SensorConfig]]

    def __init__(
        self,
        arg: Optional[
            Union[SensorConfig, dict[int, SensorConfig], list[dict[int, SensorConfig]]]
        ] = None,
        *,
        extended: Optional[bool] = None,
        update_rate: Optional[float] = None,
    ) -> None:
        self.update_rate = update_rate

        if arg is None:
            self._groups = [{1: SensorConfig()}]
        else:
            self._groups = _unsqueeze_groups(arg)
        _validate_groups_structure(self._groups)

        num_entries = sum(len(g) for g in self._groups)
        must_be_extended = num_entries > 1

        if extended is None:
            self._extended = must_be_extended
        else:
            self._extended = extended

        if extended is not None:
            if not extended and must_be_extended:
                raise ValueError("Must be extended since multiple sensor configs are given")

    @property
    def extended(self) -> bool:
        """Whether or not the extended format is used"""

        return self._extended

    @property
    def groups(self) -> list[dict[int, SensorConfig]]:
        """The sensor config groups of this session"""

        return self._groups

    @property
    def update_rate(self) -> Optional[float]:
        """The update rate limit in Hz

        Must be > 0 or None, where None means unlimited.
        """

        return self._update_rate

    @update_rate.setter
    def update_rate(self, value: Optional[float]) -> None:
        if value is None:
            self._update_rate = None
        else:
            self._update_rate = utils.validate_float(value, min_value=0.0, inclusive=False)

    def _assert_not_extended(self) -> None:
        if self.extended:
            raise RuntimeError("This operation requires SessionConfig not to be extended.")

    def _collect_validation_results(self) -> list[ValidationResult]:
        validation_results = []
        for group in self._groups:
            for _, sensor_config in group.items():
                validation_results.extend(sensor_config._collect_validation_results())

        if self.update_rate is not None:
            for group_id, sensor_id, sensor_config in utils.iterate_extended_structure(
                self._groups
            ):
                error_msg = (
                    f"Sensor config in group {group_id} with sensor id {sensor_id} "
                    + "has a set `frame_rate`. This is not allowed."
                )
                if sensor_config.frame_rate is not None:
                    validation_results.append(ValidationError(self, "update_rate", error_msg))
                    validation_results.append(
                        ValidationError(sensor_config, "frame_rate", error_msg)
                    )

        return validation_results

    def validate(self) -> None:
        """Performs self-validation and validation of its sensor configs

        :raises ValidationError: If anything is invalid.
        """
        for validation_result in self._collect_validation_results():
            try:
                raise validation_result
            except ValidationWarning as vw:
                warnings.warn(vw.message)

    @property
    def sensor_id(self) -> int:
        """Retrieves the sole sensor ID

        :raises RuntimeError: If this session config is extended
        """

        self._assert_not_extended()
        (group,) = self._groups
        (sensor_id,) = group.keys()
        return sensor_id

    @sensor_id.setter
    def sensor_id(self, sensor_id: int) -> None:
        """Sets the sole sensor ID

        :raises RuntimeError: If this session config is extended
        """

        self._assert_not_extended()
        (group,) = self._groups
        (old_sensor_id,) = group.keys()
        group[sensor_id] = group.pop(old_sensor_id)

    @property
    def sensor_config(self) -> SensorConfig:
        """Retrieves the sole sensor config

        :raises RuntimeError: If this session config is extended
        """

        self._assert_not_extended()
        (group,) = self._groups
        (sensor_config,) = group.values()
        return sensor_config

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.to_dict() == other.to_dict()

    def to_dict(self) -> dict[str, Any]:
        d: dict[str, Any] = {"groups": [], "extended": self.extended}
        for group_dict in self._groups:
            d["groups"].append(
                {
                    sensor_id: sensor_config.to_dict()
                    for sensor_id, sensor_config in group_dict.items()
                }
            )

        d["update_rate"] = self._update_rate
        return d

    @classmethod
    def from_dict(cls, d: dict) -> SessionConfig:
        d = d.copy()
        d["arg"] = []
        groups_list = d.pop("groups")

        for group_dict in groups_list:
            d["arg"].append(
                {
                    sensor_id: SensorConfig.from_dict(sensor_config_dict)
                    for sensor_id, sensor_config_dict in group_dict.items()
                }
            )

        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=utils.EntityJSONEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> SessionConfig:
        session_config_dict = json.loads(json_str)

        # sensor_ids in groups will be strings.
        groups_list = session_config_dict.pop("groups")
        session_config_dict["groups"] = []
        for group_dict in groups_list:
            session_config_dict["groups"].append(
                {
                    int(str_sensor_id): sensor_config_dict
                    for str_sensor_id, sensor_config_dict in group_dict.items()
                }
            )

        return cls.from_dict(session_config_dict)

    def __str__(self) -> str:
        lines = []

        lines.append(f"{type(self).__name__}:")

        d = self.to_dict()
        del d["groups"]
        lines.extend(utils.pretty_dict_line_strs(d))

        lines.append("  groups:")
        for group_idx, group_dict in enumerate(self.groups):
            lines.append(f"    group {group_idx}:")
            for sensor_id, sensor_config in group_dict.items():
                sc_lines = sensor_config._pretty_str_lines(sensor_id=sensor_id)
                lines.extend(utils.indent_strs(sc_lines, 3))

        return "\n".join(lines)


def _unsqueeze_groups(
    arg: Union[SensorConfig, dict[int, SensorConfig], list[dict[int, SensorConfig]]],
) -> list[dict[int, SensorConfig]]:
    if isinstance(arg, SensorConfig):
        return [{1: arg}]

    if isinstance(arg, dict):
        return [arg]

    if isinstance(arg, list):
        return arg

    raise ValueError


def _validate_groups_structure(groups: list[dict[int, SensorConfig]]) -> None:
    if len(groups) < 1:
        raise ValueError

    for group in groups:
        if len(group) < 1:
            raise ValueError

        for sensor_id, entry in group.items():
            if not isinstance(sensor_id, int):
                raise ValueError

            if not isinstance(entry, SensorConfig):
                raise ValueError



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import functools
import typing as t

import attrs
import numpy as np
import numpy.typing as npt
import typing_extensions as te

from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    Metadata,
    Result,
    ResultContext,
    SensorConfig,
)
from acconeer.exptool.a121._core.peripherals.communication.message import Message
from acconeer.exptool.a121._core.utils import map_over_extended_structure, zip3_extended_structures

from .parse_error import ParseError


class ResultInfoDict(te.TypedDict):
    tick: int
    data_saturated: bool
    frame_delayed: bool
    calibration_needed: bool
    temperature: int


class ResultMessageHeader(te.TypedDict):
    result_info: list[list[ResultInfoDict]]
    payload_size: int


@attrs.frozen
class EmptyResultMessage(Message):
    """
    This message can come from the server if it somehow starts streaming
    after being stopped.
    """

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> EmptyResultMessage:
        head = t.cast(ResultMessageHeader, header)

        if head.get("payload_size") == 0 and head.get("result_info") == []:
            return EmptyResultMessage()
        else:
            raise ParseError


@attrs.frozen
class ResultMessage(Message):
    grouped_result_infos: list[list[ResultInfoDict]]
    frame_blob: bytes

    @staticmethod
    def _create_result_context(metadata: Metadata, ticks_per_second: int) -> ResultContext:
        return ResultContext(
            metadata=metadata,
            ticks_per_second=ticks_per_second,
        )

    @staticmethod
    def _create_result(args: tuple[ResultInfoDict, npt.NDArray, ResultContext]) -> Result:
        result_info, frame, context = args
        return Result(
            **result_info,
            frame=frame,
            context=context,
        )

    @classmethod
    def _get_array_from_blob(
        cls, frame_blob: bytes, start: int, end: int, frame_shape: t.Tuple[int, int]
    ) -> npt.NDArray:
        raw_frame_data = frame_blob[start:end]
        assert len(raw_frame_data) == end - start
        np_frame = np.frombuffer(raw_frame_data, dtype=INT_16_COMPLEX)
        resized_frame = np.resize(np_frame, frame_shape)
        return resized_frame

    @classmethod
    def _divide_frame_blob(
        cls, frame_blob: bytes, extended_metadata: list[dict[int, Metadata]]
    ) -> list[dict[int, npt.NDArray]]:
        start = 0
        result = []
        for metadata_group in extended_metadata:
            result_group: dict[int, npt.NDArray] = {}
            result.append(result_group)
            for sensor_id, metadata in metadata_group.items():
                end = start + metadata.frame_data_length * 4  # 4 = sizeof(int_16_complex)

                result_group[sensor_id] = cls._get_array_from_blob(
                    frame_blob, start, end, metadata.frame_shape
                )

                start = end

        return result

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> ResultMessage:
        t.cast(ResultMessageHeader, header)
        try:
            return cls(header["result_info"], payload)
        except KeyError as ke:
            raise ParseError from ke

    def get_extended_results(
        self,
        tps: int,
        metadata: list[dict[int, Metadata]],
        config_groups: list[dict[int, SensorConfig]],
    ) -> list[dict[int, Result]]:

        extended_frames = self._divide_frame_blob(self.frame_blob, metadata)
        extended_contexts = map_over_extended_structure(
            functools.partial(self._create_result_context, ticks_per_second=tps), metadata
        )
        extended_result_infos = [
            {
                sensor_id: result_info
                for result_info, sensor_id in zip(result_info_group, config_group.keys())
            }
            for result_info_group, config_group in zip(self.grouped_result_infos, config_groups)
        ]

        extended_results = map_over_extended_structure(
            self._create_result,
            zip3_extended_structures(extended_result_infos, extended_frames, extended_contexts),
        )

        return extended_results



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool.a111.algo.breathing._meta as breathing_meta
import acconeer.exptool.a111.algo.button_press._meta as button_press_meta
import acconeer.exptool.a111.algo.button_press_sparse._meta as button_press_sparse_meta
import acconeer.exptool.a111.algo.distance_detector._meta as distance_detector_meta
import acconeer.exptool.a111.algo.envelope._meta as envelope_meta
import acconeer.exptool.a111.algo.iq._meta as iq_meta
import acconeer.exptool.a111.algo.obstacle_detection._meta as obstacle_detection_meta
import acconeer.exptool.a111.algo.parking._meta as parking_meta
import acconeer.exptool.a111.algo.phase_tracking._meta as phase_tracking_meta
import acconeer.exptool.a111.algo.power_bins._meta as power_bins_meta
import acconeer.exptool.a111.algo.presence_detection_sparse._meta as presence_detection_sparse_meta
import acconeer.exptool.a111.algo.sleep_breathing._meta as sleep_breathing_meta
import acconeer.exptool.a111.algo.sparse._meta as sparse_meta
import acconeer.exptool.a111.algo.sparse_fft._meta as sparse_fft_meta
import acconeer.exptool.a111.algo.sparse_inter_fft._meta as sparse_inter_fft_meta
import acconeer.exptool.a111.algo.speed_sparse._meta as speed_sparse_meta
import acconeer.exptool.a111.algo.tank_level_short._meta as tank_level_short
import acconeer.exptool.a111.algo.wave_to_exit._meta as wave_sparse_meta


MODULE_INFOS = [
    envelope_meta.module_info,
    iq_meta.module_info,
    power_bins_meta.module_info,
    sparse_meta.module_info,
    presence_detection_sparse_meta.module_info,
    sparse_fft_meta.module_info,
    sparse_inter_fft_meta.module_info,
    speed_sparse_meta.module_info,
    breathing_meta.module_info,
    phase_tracking_meta.module_info,
    sleep_breathing_meta.module_info,
    obstacle_detection_meta.module_info,
    button_press_meta.module_info,
    button_press_sparse_meta.module_info,
    distance_detector_meta.module_info,
    parking_meta.module_info,
    wave_sparse_meta.module_info,
    tank_level_short.module_info,
]

MODULE_KEY_TO_MODULE_INFO_MAP = {mi.key: mi for mi in MODULE_INFOS}
MODULE_LABEL_TO_MODULE_INFO_MAP = {mi.label: mi for mi in MODULE_INFOS}



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import h5py

from acconeer.exptool import a121


def _create_h5_string_dataset(group: h5py.Group, name: str, string: str) -> None:
    group.create_dataset(
        name,
        data=string,
        dtype=a121._H5PY_STR_DTYPE,
        track_times=False,
    )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt
from numpy import cos, pi, sqrt, square
from scipy.special import binom

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase, ProcessorBase
from acconeer.exptool.a121.algo._utils import get_distances_m


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    intra_enable: bool = attrs.field(default=True)
    inter_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_phase_boost: bool = attrs.field(default=False)
    phase_adaptivity_tc: float = attrs.field(default=5)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=None)
    inter_frame_fast_cutoff: float = attrs.field(default=20.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=5)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.5)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if len(config.sensor_config.subsweeps) > 1:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "subsweeps",
                    "Multiple subsweeps are not supported",
                )
            )

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.sweeps_per_frame <= Processor.NOISE_ESTIMATION_DIFF_ORDER:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweeps_per_frame",
                    f"Must be greater than {Processor.NOISE_ESTIMATION_DIFF_ORDER}",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext:
    estimated_frame_rate: Optional[float] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    frame: npt.NDArray[np.complex_] = attrs.field()
    abs_mean_sweep: npt.NDArray[np.float_] = attrs.field()
    fast_lp_mean_sweep: npt.NDArray[np.float_] = attrs.field()
    slow_lp_mean_sweep: npt.NDArray[np.float_] = attrs.field()
    lp_noise: npt.NDArray[np.float_] = attrs.field()
    presence_distance_index: int = attrs.field()


@attrs.frozen(kw_only=True)
class ProcessorResult:
    intra_presence_score: float = attrs.field()
    intra: npt.NDArray[np.float_] = attrs.field()
    inter_presence_score: float = attrs.field()
    inter: npt.NDArray[np.float_] = attrs.field()
    presence_distance: float = attrs.field()
    presence_detected: bool = attrs.field()
    extra_result: ProcessorExtraResult = attrs.field()


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    # lp(f): low pass (filtered)
    # cut: cutoff frequency [Hz]
    # tc: time constant [s]
    # sf: smoothing factor [dimensionless]

    ENVELOPE_FWHM_M = {
        a121.Profile.PROFILE_1: 0.04,
        a121.Profile.PROFILE_2: 0.07,
        a121.Profile.PROFILE_3: 0.14,
        a121.Profile.PROFILE_4: 0.19,
        a121.Profile.PROFILE_5: 0.32,
    }

    APPROX_BASE_STEP_LENGTH_M = 2.5e-3
    MAX_AMPLITUDE_WEIGHT = 15
    NOISE_ESTIMATION_DIFF_ORDER = 3

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_index: Optional[int] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        if subsweep_index is None:
            subsweep_index = 0

        if context is None:
            context = ProcessorContext()

        self.sensor_config = sensor_config
        self.metadata = metadata
        self.processor_config = processor_config
        self.subsweep_index = subsweep_index

        self.processor_config.validate(self.sensor_config)

        # Should never happen, checked in validate
        assert self.sensor_config.frame_rate is not None

        self.sweeps_per_frame = self.sensor_config.sweeps_per_frame
        self.distances, _ = get_distances_m(self.sensor_config, metadata)
        self.num_distances = self.distances.size
        if context.estimated_frame_rate is not None:
            self.f = context.estimated_frame_rate
        else:
            self.f = self.sensor_config.frame_rate

        points_per_pulse = self.ENVELOPE_FWHM_M[sensor_config.profile] / (
            self.APPROX_BASE_STEP_LENGTH_M * sensor_config.step_length
        )
        self.depth_filter_length = max(int(round(points_per_pulse)), 1)
        self.depth_filter_length = min(self.depth_filter_length, self.num_distances)

        # Fixed parameters
        self.noise_est_diff_order = self.NOISE_ESTIMATION_DIFF_ORDER
        noise_tc = 10.0

        self.noise_sf = self._tc_to_sf(noise_tc, self.f)

        nd = self.noise_est_diff_order
        self.noise_norm_factor = np.sqrt(np.sum(np.square(binom(nd, np.arange(nd + 1)))))

        self.lp_mean_sweep_for_abs = np.zeros(self.num_distances, dtype=np.complex_)
        self.lp_mean_sweep_for_phase = np.zeros(self.num_distances, dtype=np.complex_)
        self.mean_sweep_tc = self.processor_config.phase_adaptivity_tc
        self.mean_sweep_sf = self._tc_to_sf(self.mean_sweep_tc, self.f)
        self.lp_phase_shift = np.zeros(self.num_distances)
        self.inter_phase_boost = self.processor_config.inter_phase_boost

        self.inter_frame_presence_timeout = processor_config.inter_frame_presence_timeout
        self.previous_presence_score = 0
        self.negative_count = 0

        self.fast_lp_mean_sweep = np.zeros(self.num_distances)
        self.slow_lp_mean_sweep = np.zeros(self.num_distances)
        self.lp_inter_dev = np.zeros(self.num_distances)
        self.lp_intra_dev = np.zeros(self.num_distances)
        self.lp_noise = np.zeros(self.num_distances)

        self.intra_presence_score = 0
        self.inter_presence_score = 0
        self.presence_distance_index = 0
        self.presence_distance = 0

        self.update_index = 0

        self.intra_enable = processor_config.intra_enable
        self.intra_threshold = processor_config.intra_detection_threshold

        self.inter_enable = processor_config.inter_enable
        self.inter_threshold = processor_config.inter_detection_threshold

        self.update_config(processor_config)

    def update_config(self, processor_config: ProcessorConfig) -> None:
        self.intra_enable = processor_config.intra_enable
        self.inter_enable = processor_config.inter_enable
        self.intra_threshold = processor_config.intra_detection_threshold
        self.inter_threshold = processor_config.inter_detection_threshold

        self.fast_sf = self._cutoff_to_sf(processor_config.inter_frame_fast_cutoff, self.f)
        self.slow_sf = self._cutoff_to_sf(processor_config.inter_frame_slow_cutoff, self.f)
        self.inter_dev_sf = self._tc_to_sf(
            processor_config.inter_frame_deviation_time_const, self.f
        )
        self.intra_sf = self._tc_to_sf(processor_config.intra_frame_time_const, self.f)
        self.intra_output_sf = self._tc_to_sf(processor_config.intra_output_time_const, self.f)
        self.inter_output_sf = self._tc_to_sf(processor_config.inter_output_time_const, self.f)

        self.inter_frame_presence_timeout = self.processor_config.inter_frame_presence_timeout
        self.inter_phase_boost = self.processor_config.inter_phase_boost
        self.mean_sweep_tc = processor_config.phase_adaptivity_tc
        self.mean_sweep_sf = self._tc_to_sf(self.mean_sweep_tc, self.f)

    @staticmethod
    def _cutoff_to_sf(fc: float, fs: float) -> float:
        """Cutoff frequency to smoothing factor conversion"""

        if fc > 0.5 * fs:
            return 0.0

        cos_w = cos(2.0 * pi * (fc / fs))
        return 2.0 - cos_w - sqrt(square(cos_w) - 4.0 * cos_w + 3.0)  # type: ignore[no-any-return]

    @staticmethod
    def _tc_to_sf(tc: float, fs: float) -> float:
        """Time constant to smoothing factor conversion"""

        if tc <= 0.0:
            return 0.0

        return np.exp(-1.0 / (tc * fs))  # type: ignore[no-any-return]

    @staticmethod
    def _dynamic_sf(static_sf: float, update_index: int) -> float:
        return min(static_sf, 1.0 - 1.0 / (1.0 + update_index))

    @staticmethod
    def _abs_dev(
        a: npt.NDArray[np.complex_],
        axis: Optional[int] = None,
        ddof: int = 0,
        subtract_mean: bool = True,
    ) -> npt.NDArray[np.float_]:
        if subtract_mean:
            a = a - a.mean(axis=axis, keepdims=True)

        if axis is None:
            n = a.size
        else:
            n = a.shape[axis]

        if ddof < 0:
            raise ValueError("ddof must be greater than or equal to 0")

        if n <= ddof:
            raise ValueError("n must be greater than ddof")

        return np.mean(np.abs(a), axis=axis) * sqrt(n / (n - ddof))  # type: ignore[no-any-return]

    @staticmethod
    def _depth_filter(a: npt.NDArray, depth_filter_length: int) -> npt.NDArray[np.float_]:
        b = np.ones(depth_filter_length) / depth_filter_length

        return np.correlate(a, b, mode="same")

    @staticmethod
    def _calculate_phase_shift(a: npt.NDArray, b: npt.NDArray) -> npt.NDArray[np.float_]:
        phase_a = np.angle(a)
        phase_b = np.angle(b)
        phases_unwrapped = np.unwrap([phase_a, phase_b], axis=0)
        phase_shift = np.abs(phases_unwrapped[0, :] - phases_unwrapped[1, :])

        return phase_shift  # type: ignore[no-any-return]

    def _calculate_phase_and_amp_weight(self, mean_sweep: npt.NDArray) -> np.float_:
        """
        Calculation of a weight factor based on phase shift and amplitude.
        The phase shift between the mean sweep and a lp-filtered mean sweep is
        calculated to amplify slow motions.
        The amplitude of the mean sweep is multiplied with the phase shift to reduce
        noise amplification.
        Before multiplication with the phase shift, the amplitude is truncated to reduce
        side effects from very strong reflective objects.
        """

        phase_shift = self._calculate_phase_shift(self.lp_mean_sweep_for_phase, mean_sweep)
        sf = self._dynamic_sf(self.inter_dev_sf, self.update_index)
        self.lp_phase_shift = sf * self.lp_phase_shift + (1.0 - sf) * phase_shift

        self.lp_mean_sweep_for_abs = sf * self.lp_mean_sweep_for_abs + (1.0 - sf) * mean_sweep
        abs_lp_mean_sweep = np.abs(self.lp_mean_sweep_for_abs)

        norm_abs_mean_sweep = np.divide(
            abs_lp_mean_sweep,
            self.lp_noise,
            out=np.zeros_like(abs_lp_mean_sweep),
            where=(self.lp_noise > 1.0),
        )

        norm_abs_mean_sweep *= np.sqrt(self.sweeps_per_frame)

        # Truncate if amplitude is too strong
        norm_abs_mean_sweep = np.minimum(
            norm_abs_mean_sweep, np.ones(norm_abs_mean_sweep.shape[0]) * self.MAX_AMPLITUDE_WEIGHT
        )

        return self.lp_phase_shift * norm_abs_mean_sweep  # type: ignore[no-any-return]

    def _inter_presence_score_scaling(self) -> None:
        """
        Scaling of self.inter_presence_score for faster decline when loosing detection.
        Start exponential scaling after self.inter_frame_presence_timeout seconds.
        """

        if self.inter_frame_presence_timeout is None:
            raise ValueError("inter_frame_presence_timeout must be set")

        scaling_factor = np.exp(
            np.maximum(self.negative_count - self.inter_frame_presence_timeout * self.f, 0)
            / (self.inter_frame_presence_timeout * self.f)
        )
        self.inter_presence_score /= scaling_factor

    def _mean_sweep_sf_scaling(self) -> None:
        """
        Scaling of self.mean_sweep_sf for faster adaptation to the environment when
        loosing detection.
        Start exponential scaling after self.inter_frame_presence_timeout seconds.
        """

        if self.inter_frame_presence_timeout is None:
            raise ValueError("inter_frame_presence_timeout must be set")

        scaling_factor = np.exp(
            np.maximum(self.negative_count - self.inter_frame_presence_timeout * self.f, 0)
            * self.mean_sweep_tc
            / self.inter_frame_presence_timeout
        )
        self.mean_sweep_sf = self._tc_to_sf(self.mean_sweep_tc / scaling_factor, self.f)

    def process(self, result: a121.Result) -> ProcessorResult:
        frame = result.subframes[self.subsweep_index]

        # Noise estimation

        nd = self.noise_est_diff_order

        noise_diff = np.diff(frame, n=nd, axis=0)
        noise = self._abs_dev(noise_diff, axis=0, subtract_mean=False)
        noise /= self.noise_norm_factor
        sf = self._dynamic_sf(self.noise_sf, self.update_index)
        self.lp_noise = sf * self.lp_noise + (1.0 - sf) * noise

        # Intra-frame part

        sweep_dev = self._abs_dev(frame, axis=0, ddof=1)

        sf = self._dynamic_sf(self.intra_sf, self.update_index)
        self.lp_intra_dev = sf * self.lp_intra_dev + (1.0 - sf) * sweep_dev

        norm_lp_intra_dev = np.divide(
            self.lp_intra_dev,
            self.lp_noise,
            out=np.zeros(self.num_distances),
            where=(self.lp_noise > 1.0),
        )

        intra = self._depth_filter(norm_lp_intra_dev, self.depth_filter_length)

        intra_presence_distance_index = int(np.argmax(intra))
        intra_presence_distance = self.distances[intra_presence_distance_index]

        self.intra_presence_score = (
            self.intra_output_sf * self.intra_presence_score
            + (1.0 - self.intra_output_sf) * intra[intra_presence_distance_index]
        )

        # Inter-frame part

        mean_sweep = frame.mean(axis=0)
        abs_mean_sweep = np.abs(mean_sweep)

        sf = self._dynamic_sf(self.fast_sf, self.update_index)
        self.fast_lp_mean_sweep = sf * self.fast_lp_mean_sweep + (1.0 - sf) * abs_mean_sweep

        sf = self._dynamic_sf(self.slow_sf, self.update_index)
        self.slow_lp_mean_sweep = sf * self.slow_lp_mean_sweep + (1.0 - sf) * abs_mean_sweep

        inter_dev = np.abs(self.fast_lp_mean_sweep - self.slow_lp_mean_sweep)
        sf = self._dynamic_sf(self.inter_dev_sf, self.update_index)
        self.lp_inter_dev = sf * self.lp_inter_dev + (1.0 - sf) * inter_dev

        norm_lp_dev = np.divide(
            self.lp_inter_dev,
            self.lp_noise,
            out=np.zeros_like(self.lp_inter_dev),
            where=(self.lp_noise > 1.0),
        )

        norm_lp_dev *= np.sqrt(self.sweeps_per_frame)

        inter = self._depth_filter(norm_lp_dev, self.depth_filter_length)

        # Phase and amplitude weighting of inter-frame part

        if self.inter_phase_boost:
            if self.update_index == 0:
                self.lp_mean_sweep_for_phase = mean_sweep

            phase_and_amp_weight = self._calculate_phase_and_amp_weight(mean_sweep)

            inter = inter * phase_and_amp_weight

            sf = self._dynamic_sf(self.mean_sweep_sf, self.update_index)
            self.lp_mean_sweep_for_phase = (
                sf * self.lp_mean_sweep_for_phase + (1.0 - sf) * mean_sweep
            )

        inter_presence_distance_index = int(np.argmax(inter))
        inter_presence_distance = self.distances[inter_presence_distance_index]

        sf = self._dynamic_sf(self.inter_output_sf, self.update_index)
        self.inter_presence_score = (
            sf * self.inter_presence_score + (1.0 - sf) * inter[inter_presence_distance_index]
        )

        # Inter-frame presence timeout

        if self.inter_frame_presence_timeout:
            delta = self.inter_presence_score - self.previous_presence_score

            if delta < 0:
                self.negative_count += 1
            else:
                self.negative_count = 0

            self._inter_presence_score_scaling()
            self._mean_sweep_sf_scaling()

            self.previous_presence_score = self.inter_presence_score

        # Presence distance - intra presence distance is prioritized due to faster reaction time

        if self.intra_presence_score > self.intra_threshold and self.intra_enable:
            presence_detected = True
            self.presence_distance_index = intra_presence_distance_index
            self.presence_distance = intra_presence_distance
        elif self.inter_presence_score > self.inter_threshold and self.inter_enable:
            presence_detected = True
            self.presence_distance_index = inter_presence_distance_index
            self.presence_distance = inter_presence_distance
        else:
            presence_detected = False
            self.presence_distance = 0

        self.update_index += 1

        extra_result = ProcessorExtraResult(
            frame=frame,
            abs_mean_sweep=abs_mean_sweep,
            fast_lp_mean_sweep=self.fast_lp_mean_sweep,
            slow_lp_mean_sweep=self.slow_lp_mean_sweep,
            lp_noise=self.lp_noise,
            presence_distance_index=self.presence_distance_index,
        )

        return ProcessorResult(
            intra_presence_score=self.intra_presence_score,
            intra=intra,
            inter_presence_score=self.inter_presence_score,
            inter=inter,
            presence_detected=presence_detected,
            presence_distance=self.presence_distance,
            extra_result=extra_result,
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from functools import partial
from typing import Any, Generic, Mapping, Optional, TypeVar

import attrs

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from .pidgets import ParameterWidget
from .types import PidgetFactoryMapping
from .utils import VerticalGroupBox


T = TypeVar("T")


class AttrsConfigEditor(QWidget, Generic[T]):
    _config: Optional[T]

    sig_update = Signal(object)

    def __init__(
        self, title: str, factory_mapping: PidgetFactoryMapping, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)
        self._config = None
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)
        group_box = VerticalGroupBox(title, parent=self)
        self.layout().addWidget(group_box)

        self._pidget_mapping: Mapping[str, ParameterWidget] = {}

        for aspect, factory in factory_mapping.items():
            pidget = factory.create(group_box)
            pidget.sig_parameter_changed.connect(partial(self._update_config_aspect, aspect))
            group_box.layout().addWidget(pidget)

            self._pidget_mapping[aspect] = pidget

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        for _, pidget in self._pidget_mapping.items():
            pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        return unhandled_results

    def set_data(self, config: Optional[T]) -> None:
        self._config = config

    def sync(self) -> None:
        self._update_pidgets()

    def setEnabled(self, enabled: bool) -> None:
        super().setEnabled(enabled and self._config is not None)

    def _broadcast(self) -> None:
        self.sig_update.emit(self._config)

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None:
            return False

        result_handled = False

        if result.source == self._config:
            for aspect, pidget in self._pidget_mapping.items():
                if result.aspect == aspect:
                    self._pidget_mapping[result.aspect].set_note_text(
                        result.message, result.criticality
                    )
                    result_handled = True

        return result_handled

    def _update_pidgets(self) -> None:
        if self._config is None:
            return

        for aspect, pidget in self._pidget_mapping.items():
            config_value = getattr(self._config, aspect)
            pidget.set_parameter(config_value)

    def _update_config_aspect(self, aspect: str, value: Any) -> None:
        if self._config is None:
            return

        try:
            self._config = attrs.evolve(self._config, **{aspect: value})
        except Exception as e:
            self._pidget_mapping[aspect].set_note_text(e.args[0], Criticality.ERROR)
        else:
            self._pidget_mapping[aspect].set_note_text(None)

        self._broadcast()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
import logging
import queue
import shutil
import time
from enum import Enum
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union
from uuid import UUID

import attrs
from typing_extensions import Protocol

from PySide6.QtCore import QDeadlineTimer, QObject, QThread, Signal
from PySide6.QtWidgets import QApplication, QWidget

import pyqtgraph as pg

from acconeer.exptool import a121
from acconeer.exptool.app.new._enums import (
    ConnectionInterface,
    ConnectionState,
    PluginGeneration,
    PluginState,
)
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.app_model.file_detective import investigate_file
from acconeer.exptool.app.new.backend import (
    Backend,
    BackendPlugin,
    BackendPluginStateMessage,
    ClosedTask,
    ConnectionStateMessage,
    GeneralMessage,
    LogMessage,
    Message,
    PluginStateMessage,
    StatusMessage,
)
from acconeer.exptool.app.new.storage import get_config_dir, remove_temp_dir
from acconeer.exptool.utils import CommDevice, SerialDevice, USBDevice  # type: ignore[import]

from .plugin_protocols import PlotPluginInterface, ViewPluginInterface
from .port_updater import PortUpdater


log = logging.getLogger(__name__)


class PluginPresetSpec(Protocol):
    """Defines what AppModel needs to know about a plugin preset.

    Implementations are free to add additional fields.
    """

    name: str
    description: Optional[str]
    preset_id: Optional[Enum]


class PluginSpec(Protocol):
    """Defines what AppModel needs to know about a plugin.

    Implementations are free to add additional fields.
    """

    key: str
    generation: PluginGeneration
    presets: List[PluginPresetSpec]
    default_preset_id: Enum

    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        ...

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPluginInterface:
        ...

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPluginInterface:
        ...


class _BackendListeningThread(QThread):
    sig_backend_closed_task = Signal(ClosedTask)
    sig_backend_message = Signal(Message)

    def __init__(self, backend: Backend, parent: QObject) -> None:
        super().__init__(parent)
        self.backend = backend

    def run(self) -> None:
        log.debug("Backend listening thread starting...")

        while not self.isInterruptionRequested():
            try:
                item = self.backend.recv(timeout=0.1)
            except queue.Empty:
                continue

            if isinstance(item, Message):
                self.sig_backend_message.emit(item)
            elif isinstance(item, ClosedTask):
                self.sig_backend_closed_task.emit(item)
            else:
                raise AssertionError

        log.debug("Backend listening thread stopping...")


def _to_usbdevice(obj: Union[dict, USBDevice]) -> Optional[USBDevice]:
    if isinstance(obj, dict):
        return USBDevice.from_dict(obj)
    return obj


def _to_serialdevice(obj: Union[dict, SerialDevice]) -> Optional[SerialDevice]:
    if isinstance(obj, dict):
        return SerialDevice.from_dict(obj)
    return obj


@attrs.mutable(kw_only=True)
class _PersistentState:
    _FILE_NAME = "app_model_state.json"
    _ENUMS = {
        "ConnectionInterface": ConnectionInterface,
    }

    class Encoder(json.JSONEncoder):
        def default(self, obj: Any) -> Any:
            if type(obj) in _PersistentState._ENUMS.values():
                # An enum existing in _ENUMS will be transformed to:
                # {"__enum__": "RegisteredEnum.MEMBER"}
                return {"__enum__": str(obj)}
            return json.JSONEncoder.default(self, obj)

    @staticmethod
    def _to_enum(d: dict) -> Any:
        if "__enum__" in d:
            # Transform the dict {"__enum__": "RegisteredEnum.MEMBER"}
            # to the enum RegisteredEnum.MEMBER
            name, member = d["__enum__"].split(".")
            return getattr(_PersistentState._ENUMS[name], member)
        return d

    connection_interface: ConnectionInterface = attrs.field(default=ConnectionInterface.SERIAL)
    socket_connection_ip: str = attrs.field(default="")
    serial_connection_device: Optional[SerialDevice] = attrs.field(
        default=None, converter=_to_serialdevice
    )
    usb_connection_device: Optional[USBDevice] = attrs.field(default=None, converter=_to_usbdevice)
    overridden_baudrate: Optional[int] = attrs.field(default=None)
    autoconnect_enabled: bool = attrs.field(default=False)
    recording_enabled: bool = attrs.field(default=True)

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> _PersistentState:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=self.Encoder)

    @classmethod
    def from_json(cls, json_str: str) -> _PersistentState:
        return cls.from_dict(json.loads(json_str, object_hook=_PersistentState._to_enum))

    @classmethod
    def from_config_file(cls) -> _PersistentState:
        return cls.from_json((get_config_dir() / cls._FILE_NAME).read_text(encoding="utf-8"))

    def to_config_file(self) -> None:
        (get_config_dir() / _PersistentState._FILE_NAME).write_text(
            self.to_json(), encoding="utf-8"
        )


class AppModel(QObject):
    sig_notify = Signal(object)
    sig_error = Signal(Exception, object)
    sig_load_plugin = Signal(object)
    sig_message_plot_plugin = Signal(object)
    sig_message_view_plugin = Signal(object)
    sig_status_message = Signal(object)
    sig_rate_stats = Signal(float, bool, float, bool)
    sig_backend_cpu_percent = Signal(int)
    sig_frame_count = Signal(object)
    sig_backend_state_changed = Signal(object)

    plugins: list[PluginSpec]
    plugin: Optional[PluginSpec]

    backend_plugin_state: Any

    connection_state: ConnectionState
    connection_warning: Optional[str]
    plugin_state: PluginState
    available_serial_devices: List[SerialDevice]
    available_usb_devices: List[USBDevice]

    saveable_file: Optional[Path]

    def __init__(self, backend: Backend, plugins: list[PluginSpec]) -> None:
        super().__init__()
        self._backend = backend
        self._listener = _BackendListeningThread(self._backend, self)
        self._listener.sig_backend_message.connect(self._handle_backend_message)
        self._listener.sig_backend_closed_task.connect(self._handle_backend_closed_task)
        self._port_updater = PortUpdater(self)
        self._port_updater.sig_update.connect(self._handle_port_update)

        self._backend_task_callbacks: dict[UUID, Any] = {}

        self._a121_server_info: Optional[a121.ServerInfo] = None

        try:
            self._persistent_state = _PersistentState.from_config_file()
        except Exception as exc:
            if not isinstance(exc, FileNotFoundError):
                log.error("Config file loading failed, using defaults")
            self._persistent_state = _PersistentState()

        self.plugins = plugins
        self.plugin = None

        self.backend_plugin_state = None

        self.connection_state = ConnectionState.DISCONNECTED
        self.connection_warning = None
        self.plugin_state = PluginState.UNLOADED
        self._serial_connection_device = None
        self._usb_connection_device = None
        self.available_serial_devices = []
        self.available_usb_devices = []

        self.saveable_file = None

    def start(self) -> None:
        self._listener.start()
        self._port_updater.start()

    def stop(self) -> None:

        WAIT_FOR_UNLOAD_TIMEOUT = 1.0

        try:
            self._persistent_state.to_config_file()
        except Exception:
            log.error("Config saving failed")

        self.load_plugin(None)
        if self.connection_state in [ConnectionState.CONNECTING, ConnectionState.CONNECTED]:
            self.disconnect_client()

        wait_start_time = time.monotonic()
        while (
            self.plugin_state != PluginState.UNLOADED
            and self.connection_state != ConnectionState.DISCONNECTED
        ):  # TODO: Do this better
            QApplication.processEvents()

            if (time.monotonic() - wait_start_time) > WAIT_FOR_UNLOAD_TIMEOUT:
                log.error("Plugin not unloaded on stop")
                break

        remove_temp_dir()

        self._listener.requestInterruption()
        status = self._listener.wait(QDeadlineTimer(500))

        if not status:
            log.debug("Backend listening thread did not stop when requested, terminating...")
            self._listener.terminate()

        self._port_updater.stop()

    def broadcast(self) -> None:
        self.sig_notify.emit(self)

    def broadcast_backend_state(self) -> None:
        self.sig_backend_state_changed.emit(self.backend_plugin_state)

    def emit_error(self, exception: Exception, traceback_format_exc: Optional[str] = None) -> None:
        log.debug("Emitting error")
        self.sig_error.emit(exception, traceback_format_exc)

    def _put_backend_task(
        self,
        name: str,
        kwargs: Optional[dict[str, Any]] = None,
        *,
        plugin: bool = False,
        on_ok: Optional[Callable[[], None]] = None,
        on_error: Optional[Callable[[Exception, Optional[str]], None]] = None,
    ) -> None:
        if kwargs is None:
            kwargs = {}

        key = self._backend.put_task((name, kwargs, plugin))
        self._backend_task_callbacks[key] = {
            "on_ok": on_ok,
            "on_error": on_error,
        }

        log.debug(f"Put backend task with name: '{name}', key: {key.time_low}")

    def put_backend_plugin_task(
        self,
        name: str,
        kwargs: Optional[dict[str, Any]] = None,
        *,
        on_ok: Optional[Callable[[], None]] = None,
        on_error: Optional[Callable[[Exception, Optional[str]], None]] = None,
    ) -> None:
        self._put_backend_task(
            name,
            kwargs,
            plugin=True,
            on_ok=on_ok,
            on_error=on_error,
        )

    def _handle_backend_closed_task(self, closed_task: ClosedTask) -> None:
        log.debug(f"Got backend closed task: {closed_task.key.time_low}")

        callbacks = self._backend_task_callbacks.pop(closed_task.key)

        if closed_task.exception:
            f = callbacks["on_error"]
            if f:
                f(closed_task.exception, closed_task.traceback_format_exc)
        else:
            f = callbacks["on_ok"]
            if f:
                f()

    def _handle_backend_message(self, message: Message) -> None:
        if isinstance(message, ConnectionStateMessage):
            log.debug("Got backend connection state message")
            self.connection_state = message.state
            self.connection_warning = message.warning
            self.broadcast()
        elif isinstance(message, PluginStateMessage):
            log.debug("Got backend plugin state message")
            self.plugin_state = message.state
            self.broadcast()
        elif isinstance(message, BackendPluginStateMessage):
            log.debug("Got backend plugin state message")
            self.backend_plugin_state = message.state
            self.broadcast_backend_state()
            self.broadcast()
        elif isinstance(message, StatusMessage):
            self.send_status_message(message.status)
        elif isinstance(message, LogMessage):
            module_logger = logging.getLogger(message.module_name)
            loglevel_to_logfunc = {
                "CRITICAL": module_logger.critical,
                "ERROR": module_logger.error,
                "WARNING": module_logger.warning,
                "INFO": module_logger.info,
                "DEBUG": module_logger.debug,
            }
            loglevel_to_logfunc[message.log_level](message.log_string)
        elif isinstance(message, GeneralMessage):
            if message.recipient is not None:
                if message.recipient == "plot_plugin":
                    self.sig_message_plot_plugin.emit(message)
                elif message.recipient == "view_plugin":
                    self.sig_message_view_plugin.emit(message)
                else:
                    raise RuntimeError(f"Got message with unknown recipient '{message.recipient}'")
            else:
                self._handle_backend_general_message(message)
        else:
            raise RuntimeError(f"Got message of unknown type '{type(message)}'")

    def _handle_backend_general_message(self, message: GeneralMessage) -> None:
        if message.exception:
            self.emit_error(message.exception, message.traceback_format_exc)
            return

        if message.name == "server_info":
            self._a121_server_info = message.data
            self.broadcast()
        elif message.name == "saveable_file":
            assert message.data is None or isinstance(message.data, Path)
            self._update_saveable_file(message.data)
        elif message.name == "rate_stats":
            stats = message.data
            if stats is None:
                stats = a121._RateStats()
            else:
                assert isinstance(stats, a121._RateStats)

            self.sig_rate_stats.emit(
                stats.rate,
                stats.rate_warning,
                stats.jitter,
                stats.jitter_warning,
            )
        elif message.name == "cpu_percent":
            self.sig_backend_cpu_percent.emit(message.data)
        elif message.name == "frame_count":
            self.sig_frame_count.emit(message.data)
        else:
            raise RuntimeError(f"Got unknown general message '{message.name}'")

    def _update_saveable_file(self, path: Optional[Path]) -> None:
        if self.saveable_file is not None:
            self.saveable_file.unlink(missing_ok=True)

        self.saveable_file = path
        self.broadcast()

    def _is_serial_device_unflashed(self, serial_device: Optional[SerialDevice]) -> bool:
        if serial_device and serial_device.unflashed:
            return True
        return False

    def _is_usb_device_unflashed(self, usb_device: Optional[USBDevice]) -> bool:
        if usb_device and usb_device.unflashed:
            return True
        return False

    def _is_usb_device_inaccessible(self, usb_device: Optional[USBDevice]) -> bool:
        if usb_device and not usb_device.accessible:
            return True
        return False

    def _handle_port_update(
        self,
        serial_devices: list[SerialDevice],
        usb_devices: list[USBDevice],
    ) -> None:
        if self.connection_state is not ConnectionState.DISCONNECTED and (
            (
                self.connection_interface == ConnectionInterface.SERIAL
                and self.serial_connection_device not in serial_devices
            )
            or (
                self.connection_interface == ConnectionInterface.USB
                and self.usb_connection_device not in usb_devices
            )
        ):
            self.disconnect_client()

        first_update = len(self.available_usb_devices + self.available_serial_devices) == 0 and (
            self._persistent_state.usb_connection_device
            or self._persistent_state.serial_connection_device
        )

        serial_connection_device, recognized = self._select_new_device(
            self.available_serial_devices,
            serial_devices,
            self._persistent_state.serial_connection_device,
        )
        self.set_serial_connection_device(serial_connection_device)
        self.available_serial_devices = serial_devices

        connect = False

        if recognized:
            self.set_connection_interface(ConnectionInterface.SERIAL)
            if self._is_serial_device_unflashed(self.serial_connection_device):
                connect = False
                self.send_warning_message(
                    f"Found unflashed device at serial port: {self.serial_connection_device}"
                )
            else:
                connect = True
                self.send_status_message(
                    f"Recognized serial port: {self.serial_connection_device}"
                )

        usb_connection_device, recognized = self._select_new_device(
            self.available_usb_devices,
            usb_devices,
            self._persistent_state.usb_connection_device,
        )
        self.set_usb_connection_device(usb_connection_device)
        self.available_usb_devices = usb_devices

        if recognized:
            assert usb_connection_device is not None
            self.set_connection_interface(ConnectionInterface.USB)
            if self._is_usb_device_unflashed(usb_connection_device):
                connect = False
                self.send_warning_message(f"Found unflashed USB device: {usb_connection_device}")
            elif self._is_usb_device_inaccessible(usb_connection_device):
                connect = False
                self.send_warning_message(
                    f"Found inaccessible USB device: {usb_connection_device}"
                )
            else:
                connect = True
                self.send_status_message(f"Recognized USB device: {usb_connection_device}")

        if (
            (connect or first_update)
            and self.connection_state == ConnectionState.DISCONNECTED
            and self.autoconnect_enabled
        ):
            self._autoconnect()

        self.broadcast()

    def _autoconnect(self) -> None:
        self.connect_client(auto=True)

    def _select_new_device(
        self,
        old_devices: list[CommDevice],
        new_devices: list[CommDevice],
        current_port: Optional[CommDevice],
    ) -> Tuple[Optional[CommDevice], bool]:
        if self.connection_state != ConnectionState.DISCONNECTED:
            return current_port, False

        if not new_devices:
            return None, False

        added_devices = [dev for dev in new_devices if dev not in old_devices]

        # This can happen in the first port update when the current device is
        # restored from persistent state
        if current_port in added_devices:
            return current_port, False

        for device in added_devices:
            if device.recognized:
                return device, True

        if current_port not in new_devices:
            return new_devices[0], True

        return current_port, False

    def connect_client(self, auto: bool = False) -> None:
        if self.connection_interface == ConnectionInterface.SIMULATED:
            client_info = a121.ClientInfo(mock=True)
        elif self.connection_interface == ConnectionInterface.SOCKET:
            client_info = a121.ClientInfo(ip_address=self.socket_connection_ip)
        elif (
            self.connection_interface == ConnectionInterface.SERIAL
            and self.serial_connection_device is not None
        ):
            client_info = a121.ClientInfo(
                serial_port=self.serial_connection_device.port,
                override_baudrate=self.overridden_baudrate,
            )
        elif self.connection_interface == ConnectionInterface.USB:
            client_info = a121.ClientInfo(usb_device=self.usb_connection_device)
        else:
            raise RuntimeError

        log.debug(f"Connecting client with {client_info}")

        on_error = self.emit_error
        if auto:
            on_error = self._failed_autoconnect

        self._put_backend_task(
            "connect_client",
            {"client_info": client_info},
            on_error=on_error,
        )
        self.connection_state = ConnectionState.CONNECTING
        self.connection_warning = None
        self.broadcast()

    def disconnect_client(self) -> None:
        self._put_backend_task("disconnect_client", {})
        self.connection_state = ConnectionState.DISCONNECTING
        self.connection_warning = None
        self._a121_server_info = None
        self.broadcast()

    def is_connect_ready(self) -> bool:
        return (
            (self.connection_interface == ConnectionInterface.SOCKET)
            or (self.connection_interface == ConnectionInterface.SIMULATED)
            or (
                self.connection_interface == ConnectionInterface.SERIAL
                and self.serial_connection_device is not None
                and not self._is_serial_device_unflashed(self.serial_connection_device)
            )
            or (
                self.connection_interface == ConnectionInterface.USB
                and self.usb_connection_device is not None
                and not self._is_usb_device_unflashed(self.usb_connection_device)
                and not self._is_usb_device_inaccessible(self.usb_connection_device)
            )
        )

    def is_ready_for_session(self) -> bool:
        """
        Returns True if the plugin is ready for a new session.
        Additional conditions can be added in respective plugin.
        """
        return (
            self.plugin_state == PluginState.LOADED_IDLE
            and self.connection_state == ConnectionState.CONNECTED
            and bool(self.connected_sensors)
        )

    def _failed_autoconnect(
        self, exception: Exception, traceback_format_exc: Optional[str] = None
    ) -> None:
        self.send_warning_message("Failed to autoconnect")

    def set_port_updates_pause(self, pause: bool) -> None:
        if pause:
            self._port_updater.pause()
        else:
            self._port_updater.resume()

    @property
    def connection_interface(self) -> ConnectionInterface:
        return self._persistent_state.connection_interface

    @property
    def socket_connection_ip(self) -> str:
        return self._persistent_state.socket_connection_ip

    @property
    def serial_connection_device(self) -> Optional[SerialDevice]:
        return self._serial_connection_device

    @property
    def usb_connection_device(self) -> Optional[USBDevice]:
        return self._usb_connection_device

    @property
    def autoconnect_enabled(self) -> bool:
        return self._persistent_state.autoconnect_enabled

    @property
    def overridden_baudrate(self) -> Optional[int]:
        return self._persistent_state.overridden_baudrate

    @property
    def recording_enabled(self) -> bool:
        return self._persistent_state.recording_enabled

    def set_connection_interface(self, connection_interface: ConnectionInterface) -> None:
        self._persistent_state.connection_interface = connection_interface
        self.broadcast()

    def set_socket_connection_ip(self, ip: str) -> None:
        self._persistent_state.socket_connection_ip = ip
        self.broadcast()

    def set_serial_connection_device(self, device: Optional[SerialDevice]) -> None:
        self._serial_connection_device = device
        self._persistent_state.serial_connection_device = device
        self.broadcast()

    def set_usb_connection_device(self, device: Optional[USBDevice]) -> None:
        self._usb_connection_device = device
        self._persistent_state.usb_connection_device = device
        self.broadcast()

    def set_plugin_state(self, state: PluginState) -> None:
        self.plugin_state = state
        self.broadcast()

    def set_autoconnect_enabled(self, autoconnect_enabled: bool) -> None:
        self._persistent_state.autoconnect_enabled = autoconnect_enabled
        self.broadcast()

    def set_overridden_baudrate(self, overridden_baudrate: Optional[int]) -> None:
        self._persistent_state.overridden_baudrate = overridden_baudrate
        self.broadcast()

    def set_recording_enabled(self, recording_enabled: bool) -> None:
        self._persistent_state.recording_enabled = recording_enabled
        self.broadcast()

    def _unload_current_plugin(self) -> None:
        log.debug("AppModel is unloading its current plugin")
        self.sig_load_plugin.emit(None)
        self._update_saveable_file(None)
        self.backend_plugin_state = None
        if self.plugin is not None:
            self.plugin_state = PluginState.UNLOADING

        self.broadcast()

        self._put_backend_task("unload_plugin", {}, on_error=self.emit_error)

    def load_plugin(self, plugin: Optional[PluginSpec]) -> None:
        log.debug(f"AppModel is loading the plugin {plugin}")
        if plugin == self.plugin:
            return

        self._unload_current_plugin()

        if plugin is not None:
            self._put_backend_task(
                "load_plugin",
                {
                    "plugin_factory": plugin.create_backend_plugin,
                    "key": plugin.key,
                },
                on_error=self.emit_error,
            )
            self.put_backend_plugin_task("load_from_cache", {})
            self.plugin_state = PluginState.LOADING

        self.sig_load_plugin.emit(plugin)
        self.plugin = plugin
        self.broadcast()
        self.broadcast_backend_state()

    def set_plugin_preset(self, preset_id: Enum) -> None:
        self._put_backend_task(
            "set_preset",
            {
                "preset_id": preset_id.value,
            },
            plugin=True,
            on_error=self.emit_error,
        )

    def save_to_file(self, path: Path) -> None:
        log.debug(f"{self.__class__.__name__} saving to file '{path}'")

        if self.saveable_file is None:
            raise RuntimeError

        shutil.copyfile(self.saveable_file, path)

    def load_from_file(self, path: Path) -> None:
        log.debug(f"{self.__class__.__name__} loading from file '{path}'")

        findings = investigate_file(path)

        if findings is None:
            self.emit_error(HandledException("Cannot load file"))
            return

        if findings.generation != PluginGeneration.A121:
            self.emit_error(HandledException("This app can currently only load A121 files"))
            return

        try:
            plugin = self._find_plugin(findings.key)
        except Exception:
            log.debug(f"Could not find plugin '{findings.key}'")

            # TODO: Don't hardcode
            plugin = self._find_plugin("sparse_iq")  # noqa: F841

        self.load_plugin(plugin)
        self.put_backend_plugin_task("load_from_file", {"path": path}, on_error=self.emit_error)
        self.plugin_state = PluginState.LOADED_STARTING
        self.broadcast()

    def _find_plugin(self, find_key: Optional[str]) -> PluginSpec:  # TODO: Also find by generation
        if find_key is None:
            raise Exception

        return next(plugin for plugin in self.plugins if plugin.key == find_key)

    @property
    def rss_version(self) -> Optional[str]:
        if self._a121_server_info is None:
            return None

        return self._a121_server_info.rss_version

    @property
    def connected_sensors(self) -> list[int]:
        return self._a121_server_info.connected_sensors if self._a121_server_info else []

    def send_status_message(self, message: Optional[str]) -> None:
        self.sig_status_message.emit(message)

    def send_warning_message(self, message: Optional[str]) -> None:
        self.sig_status_message.emit(f'<p style="color: #FD5200;"><b>{message}</b></p>')



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._base import (
    AlgoBase,
    AlgoConfigBase,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ExtendedProcessorBase,
    GenericProcessorBase,
    ProcessorBase,
)
from ._utils import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    NOISE_TEMPERATURE_MODEL_PARAMETER,
    PERCEIVED_WAVELENGTH,
    SIGNAL_TEMPERATURE_MODEL_PARAMETER,
    find_peaks,
    get_approx_fft_vels,
    get_approx_sweep_rate,
    get_distance_filter_coeffs,
    get_distance_filter_edge_margin,
    get_distances_m,
    get_temperature_adjustment_factors,
    interpolate_peaks,
    select_prf,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import json
import logging
import time
from typing import Any, Iterator, Optional, Tuple, Type, Union

import attrs
from serial.serialutil import SerialException

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    SensorInfo,
    ServerInfo,
    ServerLogMessage,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import ClientError, Recorder
from acconeer.exptool.a121._core.utils import (
    create_extended_structure,
    iterate_extended_structure,
    unextend,
    unwrap_ticks,
)
from acconeer.exptool.a121._perf_calc import _SessionPerformanceCalc

from .common_client import CommonClient
from .communication_protocol import CommunicationProtocol
from .exploration_protocol import (
    ExplorationProtocol,
    ServerError,
    get_exploration_protocol,
    messages,
)
from .exploration_protocol.messages.system_info_response import SystemInfoDict
from .links import AdaptedSerialLink, BufferedLink, NullLinkError
from .message import Message, MessageT
from .utils import autodetermine_client_link, get_calibrations_provided, link_factory


log = logging.getLogger(__name__)


class ExplorationClient(CommonClient):
    _link: BufferedLink
    _default_link_timeout: float
    _link_timeout: float
    _protocol: Type[CommunicationProtocol]
    _protocol_overridden: bool
    _tick_unwrapper: TickUnwrapper
    _sensor_infos: dict[int, SensorInfo]
    _system_info: Optional[SystemInfoDict]
    _result_queue: list[list[dict[int, Result]]]
    _message_stream: Iterator[Message]
    _log_queue: list[ServerLogMessage]

    def __init__(
        self,
        client_info: ClientInfo,
        _override_protocol: Optional[Type[CommunicationProtocol]] = None,
    ) -> None:
        super().__init__(client_info)
        self._link = link_factory(self.client_info)
        self._tick_unwrapper = TickUnwrapper()
        self._message_stream = iter([])
        self._sensor_infos = {}
        self._system_info = None
        self._log_queue = []

        self._protocol: Type[CommunicationProtocol] = ExplorationProtocol
        self._protocol_overridden = False

        if _override_protocol is not None:
            self._protocol = _override_protocol
            self._protocol_overridden = True

    def _assert_connected(self) -> None:
        if not self.connected:
            raise ClientError("Client is not connected.")

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            raise ClientError("Session is not started.")

    def _get_message_stream(self) -> Iterator[Message]:
        """returns an iterator of parsed messages"""
        while True:
            header: dict[str, Any] = json.loads(self._link.recv_until(self._protocol.end_sequence))
            try:
                payload_size = header["payload_size"]
            except KeyError:
                payload = bytes()
            else:
                payload = self._link.recv(payload_size)

            resp = self._protocol.parse_message(header, payload)
            yield resp

    def _assert_deadline_not_reached(self, deadline: Optional[float]) -> None:
        if deadline is not None and time.monotonic() > deadline:
            raise ClientError("Client timed out.")

    def _apply_messages_until_message_type_encountered(
        self, message_type: Type[MessageT], timeout_s: Optional[float] = None
    ) -> MessageT:
        """Retrieves and applies messages until a message of type ``message_type`` is encountered.

        :param message_type: a subclass of ``Message``
        :param timeout_s: Limit the time spent in this function
        :raises ClientError:
            if timeout_s is set and that amount of time has elapsed
            without predicate evaluating to True
        """
        deadline = None if (timeout_s is None) else time.monotonic() + timeout_s

        if message_type in [messages.ErroneousMessage, messages.EmptyResultMessage]:
            raise ClientError("Cannot wait for error messages")

        for message in self._message_stream:
            if type(message) == messages.LogMessage:
                self._log_queue.append(message.message)
            elif type(message) == messages.EmptyResultMessage:
                raise RuntimeError("Received an empty Result from Server.")
            elif type(message) == messages.ErroneousMessage:
                last_error = ""
                for log in self._log_queue:
                    if log.level == "ERROR" and "exploration_server" not in log.module:
                        last_error = f" ({log.log})"
                raise ServerError(f"{message}{last_error}")

            if type(message) == message_type:
                return message

            self._assert_deadline_not_reached(deadline)
        raise ClientError("No message received")

    def _connect_link(self) -> None:
        self._default_link_timeout = self._link.timeout
        self._link_timeout = self._default_link_timeout

        try:
            self._link.connect()
        except SerialException as exc:
            if "Permission denied" in str(exc):
                text = "\n".join(
                    [
                        str(exc),
                        "",
                        "You are probably missing permissions to access the serial port.",
                        "",
                        "Run the setup script to fix it:",
                        "$ python -m acconeer.exptool.setup",
                        "",
                        "Reboot for the changes to take effect.",
                    ]
                )
                raise ClientError(text) from exc
            else:
                raise

        self._message_stream = self._get_message_stream()

        self._link.send(self._protocol.get_system_info_command())
        system_info_response = self._apply_messages_until_message_type_encountered(
            messages.SystemInfoResponse
        )
        self._system_info = system_info_response.system_info

        self._link.send(self._protocol.get_sensor_info_command())
        sensor_info_response = self._apply_messages_until_message_type_encountered(
            messages.SensorInfoResponse
        )
        self._sensor_infos = sensor_info_response.sensor_infos

        sensor = self._system_info.get("sensor") if self._system_info else None
        if sensor != "a121":
            self._link.disconnect()
            raise ClientError(f"Wrong sensor version, expected a121 but got {sensor}")

        if not self._protocol_overridden:
            self._update_protocol_based_on_servers_rss_version()

        self._update_baudrate()

    def _update_protocol_based_on_servers_rss_version(self) -> None:
        try:
            new_protocol = get_exploration_protocol(self.server_info.parsed_rss_version)
        except Exception:
            self.disconnect()
            raise
        else:
            self._protocol = new_protocol

    def _update_baudrate(self) -> None:
        # Only Change baudrate for AdaptedSerialLink
        if not isinstance(self._link, AdaptedSerialLink):
            return

        DEFAULT_BAUDRATE = 115200
        overridden_baudrate = self.client_info.override_baudrate
        max_baudrate = self.server_info.max_baudrate
        baudrate_to_use = self.server_info.max_baudrate or DEFAULT_BAUDRATE

        # Override baudrate?
        if overridden_baudrate is not None and max_baudrate is not None:
            # Valid Baudrate?
            if overridden_baudrate > max_baudrate:
                raise ClientError(f"Cannot set a baudrate higher than {max_baudrate}")
            elif overridden_baudrate < DEFAULT_BAUDRATE:
                raise ClientError(f"Cannot set a baudrate lower than {DEFAULT_BAUDRATE}")
            baudrate_to_use = overridden_baudrate

        # Do not change baudrate if DEFAULT_BAUDRATE
        if baudrate_to_use == DEFAULT_BAUDRATE:
            return

        self._link.send(self._protocol.set_baudrate_command(baudrate_to_use))

        self._apply_messages_until_message_type_encountered(messages.SetBaudrateResponse)
        self._baudrate_ack_received = False

        self._link.baudrate = baudrate_to_use

    def connect(self) -> None:
        try:
            self._connect_link()
        except NullLinkError:
            self._client_info = autodetermine_client_link(self.client_info)
            self._link = link_factory(self.client_info)
            self._connect_link()

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        if not self.connected:
            self.connect()

        if self.session_is_started:
            raise ClientError("Session is currently running, can't setup.")

        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        config.validate()

        self._calibrations_provided = get_calibrations_provided(config, calibrations)

        self._link.send(self._protocol.setup_command(config, calibrations))

        self._session_config = config

        message = self._apply_messages_until_message_type_encountered(messages.SetupResponse)
        self._metadata = [
            {
                sensor_id: metadata
                for metadata, sensor_id in zip(metadata_group, config_group.keys())
            }
            for metadata_group, config_group in zip(
                message.grouped_metadatas, self._session_config.groups
            )
        ]
        self._sensor_calibrations = message.sensor_calibrations
        pc = _SessionPerformanceCalc(config, self._metadata)

        try:
            # Use max of the calculate duration and update/frame rate to guarantee sufficient
            # timeout.
            timeout_duration = max(pc.update_duration, 1 / pc.update_rate)
            # Increase timeout if update rate is very low, otherwise keep default
            self._link_timeout = max(1.5 * timeout_duration + 1.0, self._default_link_timeout)
        except Exception:
            self._link_timeout = self._default_link_timeout

        self._link.timeout = self._link_timeout

        if self.session_config.extended:
            return self._metadata
        else:
            return unextend(self._metadata)

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        self._assert_session_setup()

        if self.session_is_started:
            raise ClientError("Session is already started.")

        self._recorder_start(recorder)
        self._create_rate_stats_calc()
        self._session_is_started = True

        self._link.timeout = self._link_timeout
        self._link.send(self._protocol.start_streaming_command())
        self._apply_messages_until_message_type_encountered(messages.StartStreamingResponse)

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        self._assert_session_started()

        result_message = self._apply_messages_until_message_type_encountered(
            messages.ResultMessage
        )

        if self._metadata is None:
            raise RuntimeError(f"{self} has no metadata")

        if self._system_info is None:
            raise RuntimeError(f"{self} has no system info")

        if self._session_config is None:
            raise RuntimeError(f"{self} has no session config")

        extended_results = result_message.get_extended_results(
            tps=self._system_info["ticks_per_second"],
            metadata=self._metadata,
            config_groups=self._session_config.groups,
        )

        extended_results = self._tick_unwrapper.unwrap_ticks(extended_results)

        self._recorder_sample(extended_results)
        self._update_rate_stats_calc(extended_results)
        return self._return_results(extended_results)

    def stop_session(self) -> Any:
        self._assert_session_started()

        self._link.send(self._protocol.stop_streaming_command())

        try:
            self._apply_messages_until_message_type_encountered(
                messages.StopStreamingResponse, timeout_s=self._link.timeout + 1
            )
            self._session_is_started = False
        except Exception:
            raise
        finally:
            self._link.timeout = self._default_link_timeout
            self._tick_unwrapper = TickUnwrapper()

        recorder_result = self._recorder_stop()
        self._rate_stats_calc = None
        self._log_queue.clear()

        return recorder_result

    def disconnect(self) -> None:
        # TODO: Make sure this cleans up corner-cases (like lost connection)
        #       to not hog resources.

        if self.session_is_started:
            _ = self.stop_session()

        self._system_info = None
        self._sensor_infos = {}
        self._message_stream = iter([])
        self._metadata = None
        self._log_queue.clear()
        self._link.disconnect()

    @property
    def connected(self) -> bool:
        return self._sensor_infos != {} and self._system_info is not None

    @property
    def server_info(self) -> ServerInfo:
        self._assert_connected()

        assert self._system_info is not None  # Should never happend if client is connected
        return ServerInfo(
            rss_version=self._system_info["rss_version"],
            sensor_count=self._system_info["sensor_count"],
            ticks_per_second=self._system_info["ticks_per_second"],
            hardware_name=self._system_info.get("hw", None),
            sensor_infos=self._sensor_infos,
            max_baudrate=self._system_info.get("max_baudrate"),
        )


class TickUnwrapper:
    """Wraps unwrap_ticks to be applied over extended results"""

    def __init__(self) -> None:
        self.next_minimum_tick: Optional[int] = None

    def unwrap_ticks(self, extended_results: list[dict[int, Result]]) -> list[dict[int, Result]]:
        result_items = list(iterate_extended_structure(extended_results))
        ticks = [result.tick for _, _, result in result_items]
        unwrapped_ticks, self.next_minimum_tick = unwrap_ticks(ticks, self.next_minimum_tick)

        def f(result_item: Tuple[int, int, Result], updated_tick: int) -> Tuple[int, int, Result]:
            group_index, sensor_id, result = result_item
            updated_result = attrs.evolve(result, tick=updated_tick)
            return (group_index, sensor_id, updated_result)

        return create_extended_structure(map(f, result_items, unwrapped_ticks))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Iterator, Optional

import attrs

from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Record,
    RecordException,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
    StackedResults,
)


class InMemoryRecordException(Exception):
    pass


@attrs.frozen(kw_only=True)
class InMemoryRecord(Record):
    _client_info: ClientInfo = attrs.field()
    _extended_metadata: list[dict[int, Metadata]] = attrs.field()
    _extended_stacked_results: list[dict[int, StackedResults]] = attrs.field()
    _lib_version: str = attrs.field()
    _num_frames: int = attrs.field()
    _server_info: ServerInfo = attrs.field()
    _session_config: SessionConfig = attrs.field()
    _timestamp: str = attrs.field()
    _uuid: str = attrs.field()
    _calibrations: Optional[dict[int, SensorCalibration]] = attrs.field()
    _calibrations_provided: Optional[dict[int, bool]] = attrs.field()

    @property
    def client_info(self) -> ClientInfo:
        return self._client_info

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        return self._extended_metadata

    @property
    def extended_results(self) -> Iterator[list[dict[int, Result]]]:
        for frame_no in range(self.num_frames):
            yield self._get_result_for_all_entries(frame_no)

    def _get_result_for_all_entries(self, frame_no: int) -> list[dict[int, Result]]:
        def stacked_results_to_result(stacked_results: StackedResults) -> Result:
            return stacked_results[frame_no]

        return utils.map_over_extended_structure(
            stacked_results_to_result, self._extended_stacked_results
        )

    @property
    def extended_stacked_results(self) -> list[dict[int, StackedResults]]:
        return self._extended_stacked_results

    @property
    def lib_version(self) -> str:
        return self._lib_version

    @property
    def num_frames(self) -> int:
        return self._num_frames

    @property
    def server_info(self) -> ServerInfo:
        return self._server_info

    @property
    def session_config(self) -> SessionConfig:
        return self._session_config

    @property
    def sensor_id(self) -> int:
        return self._session_config.sensor_id

    @property
    def timestamp(self) -> str:
        return self._timestamp

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        if self._calibrations is None:
            raise InMemoryRecordException("No calibration in record")
        return self._calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        if self._calibrations_provided is None:
            raise InMemoryRecordException("No calibration in record")
        return self._calibrations_provided

    @classmethod
    def from_record(cls, record: Record) -> InMemoryRecord:
        try:
            calibrations = record.calibrations
            calibrations_provided = record.calibrations_provided
        except RecordException:
            calibrations = None
            calibrations_provided = None

        return cls(
            client_info=record.client_info,
            extended_metadata=record.extended_metadata,
            extended_stacked_results=record.extended_stacked_results,
            lib_version=record.lib_version,
            num_frames=record.num_frames,
            server_info=record.server_info,
            session_config=record.session_config,
            timestamp=record.timestamp,
            uuid=record.uuid,
            calibrations=calibrations,
            calibrations_provided=calibrations_provided,
        )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attr import Attribute

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities import Result
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._core.utils import is_divisor_of, is_multiple_of
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import ENVELOPE_FWHM_M, AlgoConfigBase, select_prf

from ._processors import Processor, ProcessorConfig, ProcessorContext, ProcessorExtraResult


SPARSE_IQ_PPC = 24


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=1.0)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=2.0)
    """End point of measurement interval in meters."""

    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    """
    Sets the profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR.
    """

    step_length: Optional[int] = attrs.field(default=None)
    """
    Step length in points. If no argument is provided, the step length is automatically
    calculated based on the profile.
    """

    frame_rate: float = attrs.field(default=10.0)
    """Frame rate in Hz."""

    sweeps_per_frame: int = attrs.field(default=16)
    """Number of sweeps per frame."""

    hwaas: int = attrs.field(default=32)
    """Number of HWAAS."""

    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.DEEP_SLEEP, converter=idle_state_converter
    )
    """Sets the inter frame idle state."""

    intra_enable: bool = attrs.field(default=True)
    """
    Enables the intra-frame presence detection used for detecting
    faster movements inside frames.
    """

    intra_detection_threshold: float = attrs.field(default=1.3)
    """Detection threshold for the intra-frame presence detection."""

    intra_frame_time_const: float = attrs.field(default=0.15)
    """Time constant for the depthwise filtering in the intra-frame part."""

    intra_output_time_const: float = attrs.field(default=0.5)
    """Time constant for the output in the intra-frame part."""

    inter_enable: bool = attrs.field(default=True)
    """
    Enables the inter-frame presence detection used for detecting
    slower movements between frames
    """

    inter_detection_threshold: float = attrs.field(default=1)
    """Detection threshold for the inter-frame presence detection."""

    inter_frame_fast_cutoff: float = attrs.field(default=20.0)
    """
    Cutoff frequency of the low pass filter for the fast filtered absolute sweep mean.
    No filtering is applied if the cutoff is set over half the frame rate (Nyquist limit).
    """

    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    """Cutoff frequency of the low pass filter for the slow filtered absolute sweep mean."""

    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    """Time constant of the low pass filter for the inter-frame deviation between fast and slow."""

    inter_output_time_const: float = attrs.field(default=5)
    """Time constant for the output in the inter-frame part."""

    inter_phase_boost: bool = attrs.field(default=False)
    """Enables the inter-frame phase boost. Used to increase slow motion detection."""

    inter_frame_presence_timeout: Optional[int] = attrs.field(default=None)
    """
    Number of seconds the inter-frame presence score needs to decrease before exponential
    scaling starts for faster decline.
    """

    @step_length.validator
    def _validate_step_length(self, attrs: Attribute, step_length: int) -> None:
        if step_length is not None:
            if not (
                is_divisor_of(SPARSE_IQ_PPC, step_length)
                or is_multiple_of(SPARSE_IQ_PPC, step_length)
            ):
                raise ValueError(f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}")

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.sweeps_per_frame <= Processor.NOISE_ESTIMATION_DIFF_ORDER:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweeps_per_frame",
                    f"Must be greater than {Processor.NOISE_ESTIMATION_DIFF_ORDER}",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class DetectorResult:
    intra_presence_score: float = attrs.field()
    """A measure of the amount of fast motion detected."""

    intra_depthwise_scores: npt.NDArray[np.float_] = attrs.field()
    """The depthwise presence scores for fast motions"""

    inter_presence_score: float = attrs.field()
    """A measure of the amount of slow motion detected"""

    inter_depthwise_scores: npt.NDArray[np.float_] = attrs.field()
    """The depthwise presence scores for slow motions"""

    presence_distance: float = attrs.field()
    """The distance, in meters, to the detected object"""

    presence_detected: bool = attrs.field()
    """True if presence was detected, False otherwise"""

    processor_extra_result: ProcessorExtraResult = attrs.field()
    service_result: a121.Result = attrs.field()


class Detector:
    MIN_DIST_M = {
        a121.Profile.PROFILE_1: None,
        a121.Profile.PROFILE_2: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_2],
        a121.Profile.PROFILE_3: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_3],
        a121.Profile.PROFILE_4: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_4],
        a121.Profile.PROFILE_5: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_5],
    }

    def __init__(
        self,
        *,
        client: a121.ClientBase,
        sensor_id: int,
        detector_config: DetectorConfig,
    ) -> None:
        self.client = client
        self.sensor_id = sensor_id
        self.detector_config = detector_config

        self.started = False

    def _estimate_frame_rate(self) -> float:
        delta_times = np.full(2, np.nan)

        self.client.setup_session(self.session_config)
        self.client.start_session()

        for i in range(4):
            result = self.client.get_next()
            assert isinstance(result, Result)

            if i < 2:
                last_time = result.tick_time
                continue

            time = result.tick_time
            delta = time - last_time
            last_time = time
            delta_times = np.roll(delta_times, -1)
            delta_times[-1] = delta

        self.client.stop_session()

        return float(1.0 / np.nanmean(delta_times))

    def start(self, recorder: Optional[a121.Recorder] = None) -> None:
        if self.started:
            raise RuntimeError("Already started")

        sensor_config = self._get_sensor_config(self.detector_config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        self.estimated_frame_rate = self._estimate_frame_rate()
        # Add estimated frame rate to context if it differs more than 10% from the set frame rate
        if (
            np.abs(self.detector_config.frame_rate - self.estimated_frame_rate)
            / self.detector_config.frame_rate
            > 0.1
        ):
            context = ProcessorContext(estimated_frame_rate=self.estimated_frame_rate)
        else:
            context = ProcessorContext(estimated_frame_rate=None)

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        processor_config = self._get_processor_config(self.detector_config)

        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=processor_config,
            context=context,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("presence_detector")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.detector_config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self.client.start_session(recorder)

        self.started = True

    @classmethod
    def _get_sensor_config(cls, detector_config: DetectorConfig) -> a121.SensorConfig:
        start_point = int(np.floor(detector_config.start_m / Processor.APPROX_BASE_STEP_LENGTH_M))
        if detector_config.profile is not None:
            profile = detector_config.profile
        else:
            viable_profiles = [
                k for k, v in cls.MIN_DIST_M.items() if v is None or v <= detector_config.start_m
            ]
            profile = viable_profiles[-1]

        if detector_config.step_length is not None:
            step_length = detector_config.step_length
        else:
            # Calculate biggest possible step length based on the fwhm of the set profile
            # Achieve detection on the complete range with minimum number of sampling points
            fwhm_p = Processor.ENVELOPE_FWHM_M[profile] / Processor.APPROX_BASE_STEP_LENGTH_M
            if fwhm_p < SPARSE_IQ_PPC:
                step_length = SPARSE_IQ_PPC // int(np.ceil(SPARSE_IQ_PPC / fwhm_p))
            else:
                step_length = int((fwhm_p // SPARSE_IQ_PPC) * SPARSE_IQ_PPC)

        num_point = int(
            np.ceil(
                (detector_config.end_m - detector_config.start_m)
                / (step_length * Processor.APPROX_BASE_STEP_LENGTH_M)
            )
            + 1
        )
        end_point = start_point + (num_point - 1) * step_length
        return a121.SensorConfig(
            profile=profile,
            start_point=start_point,
            num_points=num_point,
            step_length=step_length,
            prf=select_prf(end_point, profile),
            hwaas=detector_config.hwaas,
            sweeps_per_frame=detector_config.sweeps_per_frame,
            frame_rate=detector_config.frame_rate,
            inter_frame_idle_state=detector_config.inter_frame_idle_state,
        )

    @classmethod
    def _get_processor_config(cls, detector_config: DetectorConfig) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=detector_config.intra_enable,
            intra_detection_threshold=detector_config.intra_detection_threshold,
            intra_frame_time_const=detector_config.intra_frame_time_const,
            intra_output_time_const=detector_config.intra_output_time_const,
            inter_enable=detector_config.inter_enable,
            inter_detection_threshold=detector_config.inter_detection_threshold,
            inter_frame_fast_cutoff=detector_config.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=detector_config.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=detector_config.inter_frame_deviation_time_const,
            inter_output_time_const=detector_config.inter_output_time_const,
            inter_phase_boost=detector_config.inter_phase_boost,
            inter_frame_presence_timeout=detector_config.inter_frame_presence_timeout,
        )

    def get_next(self) -> DetectorResult:
        if not self.started:
            raise RuntimeError("Not started")

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return DetectorResult(
            intra_presence_score=processor_result.intra_presence_score,
            intra_depthwise_scores=processor_result.intra,
            inter_presence_score=processor_result.inter_presence_score,
            inter_depthwise_scores=processor_result.inter,
            presence_distance=processor_result.presence_distance,
            presence_detected=processor_result.presence_detected,
            processor_extra_result=processor_result.extra_result,
            service_result=result,
        )

    def update_config(self, config: DetectorConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        recorder_result = self.client.stop_session()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    detector_config: DetectorConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", detector_config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, DetectorConfig]:
    sensor_id = algo_group["sensor_id"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])
    return sensor_id, config
